class Stitch(object):
    def __init__(self):
        pass

    def __call__(self, topology, arcs, *args, **kwargs):
        self.stitched_arcs = dict()
        self.fragment_by_start = dict()
        self.fragment_by_end = dict()
        self.fragments = list()
        self.empty_index = -1
        self.topology = topology

        # Stitch empty arcs first, since they may be subsumed by other arcs.
        for i, j in zip(arcs[0::2], arcs[1::2]):
            idx = self._get_idx(i)
            arc = self.topology["arcs"][idx]
            if len(arc) < 3 and not arc[1][0] and not arc[1][1]:
                self.empty_index += 1
                arcs[self.empty_index], arcs[j] = i, arcs[self.empty_index]

        for i in arcs:
            e = self.ends(i)
            start, end = tuple(e[0]), tuple(e[1])

            if start in self.fragment_by_end:
                f = self.fragment_by_end[start]
                self.fragment_by_end.pop(f["end"], None)
                idx = (
                    sorted(list(filter(lambda x: isinstance(x, int), f.keys())))[-1] + 1
                )
                f[idx] = i
                f["end"] = end

                # g = self.fragment_by_start.get(end, None)
                # if g is not None:
                if end in self.fragment_by_start:
                    g = self.fragment_by_start[end]
                    self.fragment_by_start.pop(g["start"], None)
                    fg = f if g == f else f + g
                    fg["start"] = f["start"]
                    fg["end"] = g["end"]
                    self.fragment_by_start[fg["start"]] = fg
                    self.fragment_by_end[fg["end"]] = fg
                else:
                    self.fragment_by_start[f["start"]] = f
                    self.fragment_by_end[f["end"]] = f
            else:
                if end in self.fragment_by_start:
                    f = self.fragment_by_start[end]
                    self.fragment_by_start.pop(f["start"], None)
                    self.unshift(f, i)
                    f["start"] = start
                    g = self.fragment_by_end.get(start, None)
                    if g is not None:
                        self.fragment_by_end.pop(g["end"], None)
                        gf = f if g == f else g + f
                        gf["start"] = g["start"]
                        gf["end"] = f["end"]
                        self.fragment_by_start[gf["start"]] = gf
                        self.fragment_by_end[gf["end"]] = gf
                    else:
                        self.fragment_by_start[f["start"]] = f
                        self.fragment_by_end[f["end"]] = f
                else:
                    f = {0: i, "start": tuple(start), "end": tuple(end)}
                    self.fragment_by_start[f["start"]] = f
                    self.fragment_by_end[f["end"]] = f

        self.flush(self.fragment_by_end, self.fragment_by_start)
        self.flush(self.fragment_by_start, self.fragment_by_end)

        for i in arcs:
            idx = self._get_idx(i)
            if idx not in self.stitched_arcs:
                self.fragments.append([i])

        # Convert dict to list
        fragments = list()
        for f in self.fragments:
            if 1 in f:
                fragments.append([f[0], f[1]])
            else:
                fragments.append([f[0]])

        return fragments

    @staticmethod
    def _get_idx(i):
        return ~i if i < 0 else i

    @staticmethod
    def unshift(d, value):
        sorted_keys = sorted(
            list(filter(lambda x: isinstance(x, int), d.keys())), reverse=True
        )
        for k in sorted_keys:
            d[k + 1] = d[k]
        d[0] = value

    def ends(self, i):
        idx = self._get_idx(i)
        arc = self.topology["arcs"][idx]
        p_0 = arc[0]

        if "transform" in self.topology:
            p_1 = [0, 0]
            for dp in arc:
                p_1[0] += dp[0]
                p_1[1] += dp[1]
        else:
            p_1 = arc[-1]

        return [p_1, p_0] if i < 0 else [p_0, p_1]

    def flush(self, fragment_by_end, fragment_by_start):
        for _, f in fragment_by_end.items():
            fragment_by_start.pop(f["start"], None)
            f.pop("start", None)
            f.pop("end", None)
            for _, i in f.items():
                self.stitched_arcs[self._get_idx(i)] = 1
            self.fragments.append(f)



from pytopojson import bisect


class Neighbors(object):
    def __init__(self):
        self.bisect = bisect.Bisect()
        self.indexes_by_arc = None
        self.neighbors = None
        self.geometry_type = {
            "LineString": self.line,
            "MultiLineString": self.polygon,
            "Polygon": self.polygon,
            "MultiPolygon": self.multipolygon,
        }

    def __call__(self, objects, *args, **kwargs):
        self.indexes_by_arc = dict()  # Arc index -> array of object indexes
        self.neighbors = [[] for _ in objects]

        for idx, o in enumerate(objects):
            self.geometry(o, idx)

        for i in self.indexes_by_arc:
            indexes = self.indexes_by_arc[i]
            for j in range(len(indexes)):
                for k in range(j + 1, len(indexes)):
                    ij = indexes[j]
                    ik = indexes[k]

                    n = self.neighbors[ij]
                    i = self.bisect(n, ik)
                    if i >= len(n) or n[i] != ik:
                        self.neighbors[ij] = n[:i] + [ik] + n[i:]

                    n = self.neighbors[ik]
                    i = self.bisect(n, ij)
                    if i >= len(n) or n[i] != ij:
                        self.neighbors[ik] = n[:i] + [ij] + n[i:]

        return self.neighbors

    def line(self, arcs, i):
        for a in arcs:
            if a < 0:
                a = ~a

            if a in self.indexes_by_arc:
                self.indexes_by_arc[a].append(i)
            else:
                self.indexes_by_arc[a] = [i]

    def polygon(self, arcs, i):
        for arc in arcs:
            self.line(arc, i)

    def multipolygon(self, arcs, i):
        for arc in arcs:
            self.polygon(arc, i)

    def geometry(self, o, i):
        if o["type"] == "GeometryCollection":
            for geom in o["geometries"]:
                self.geometry(geom, i)
        elif o["type"] in self.geometry_type.keys():
            self.geometry_type[o["type"]](o["arcs"], i)



import struct


def to_bin(value):
    if type(value) == str:
        b = "".join(map(bin, bytearray(value))).replace("b", "")[:64]
        return b if len(b) == 64 else b.zfill(64)
    else:
        return float_to_bin(value)


def float_to_bin(value):
    """ Convert float to 64-bit binary string. """
    [d] = struct.unpack(">Q", struct.pack(">d", value))
    return "{:064b}".format(d)


def bin_to_uint(value):
    return int(value, 2)


def hash(point):
    point = list(point.items()[0]) if type(point) == dict else point
    bins = [to_bin(point[0]), to_bin(point[1])]
    uints = [
        bin_to_uint(bins[0][:32]),
        bin_to_uint(bins[0][32:]),
        bin_to_uint(bins[1][:32]),
        bin_to_uint(bins[1][32:]),
    ]
    h = uints[0] ^ uints[1]
    h = h << 5 ^ h >> 7 ^ uints[2] ^ uints[3]
    return h & 0x7FFFFFFF


def equal(p_1, p_2):
    return p_1 == p_2



import numpy as np


class Array(object):
    def __init__(self, size=None):
        if size is not None:
            self.list = [0] * size
            self.size = size
        else:
            self.list = []
            self.size = 0

    def __getitem__(self, key):
        if isinstance(key, slice) or key < self.size:
            return self.list[key]
        else:
            return None

    def __setitem__(self, key, value):
        if key < self.size:
            self.list[key] = self._convert(value)

    def __len__(self):
        return self.size

    def __repr__(self):
        return "{classname}({list_obj})".format(
            classname=self.__class__.__name__, list_obj=self.list.__repr__()
        )

    def __str__(self):
        return "{classname}({list_obj})".format(
            classname=self.__class__.__name__, list_obj=self.list.__str__()
        )

    @staticmethod
    def _convert(value):
        return value


class Int8Array(Array):
    def __init__(self, size=None):
        super(Int8Array, self).__init__(size)

    @staticmethod
    def _convert(value):
        return np.array([value]).astype(np.int8)[0]


class Int16Array(Array):
    def __init__(self, size=None):
        super(Int16Array, self).__init__(size)

    @staticmethod
    def _convert(value):
        return np.array([value]).astype(np.int16)[0]


class Int32Array(Array):
    def __init__(self, size=None):
        super(Int32Array, self).__init__(size)

    @staticmethod
    def _convert(value):
        return np.array([value]).astype(np.int32)[0]


class ListDict(dict):
    def __init__(self, *args, **kw):
        super(ListDict, self).__init__(*args, **kw)

    def keys(self):
        return sorted(list(super(ListDict, self).keys()))

    def values(self):
        vals = list()
        for k in self.keys():
            vals.append(self[k])
        return vals

    def append(self, value):
        k = len(self.keys())
        self[k] = value

    def __delitem__(self, key):
        value = super().pop(key)
        super().pop(value, None)

    def __setitem__(self, key, value):
        if key in self:
            del self[key]
        super().__setitem__(key, value)

    def __iter__(self):
        self.index = 0
        return self

    def __next__(self):
        try:
            x = self[self.index]
        except KeyError:
            raise StopIteration
        self.index += 1
        return x

    def __str__(self):
        return str(self.values())

    def __unicode__(self):
        return str(self.values())

    def __repr__(self):
        return str(self.values())

    def to_list(self):
        return self.values()


if __name__ == "__main__":
    a = Int8Array(5)
    print(a)
    a[3] = 2387
    print(a)
    a[10] = 10
    print(a)

    a = ListDict({0: 1, 1: 2})
    a.append(123)
    print(a)



import unittest

from pytopojson import prequantize


class PrequantizeTestCase(unittest.TestCase):
    def setUp(self):
        self.prequantize = prequantize.Prequantize()

    def test_prequantize_returns_the_quantization_transform(self):
        quantize = self.prequantize({}, [0, 0, 1, 1], 1e4)

        self.assertDictEqual(
            quantize, {"scale": [1 / 9999, 1 / 9999], "translate": [0, 0]}
        )

    def test_prequantize_converts_coordinates_to_fixed_precision(self):
        objects = {
            "foo": {"type": "LineString", "arcs": [[0, 0], [1, 0], [0, 1], [0, 0]]}
        }

        self.prequantize(objects, [0, 0, 1, 1], 1e4)
        self.assertCountEqual(
            objects["foo"]["arcs"], [[0, 0], [9999, 0], [0, 9999], [0, 0]]
        )

    def test_prequantize_observes_the_quantization_parameter(self):
        objects = {
            "foo": {"type": "LineString", "arcs": [[0, 0], [1, 0], [0, 1], [0, 0]]}
        }

        self.prequantize(objects, [0, 0, 1, 1], 10)
        self.assertCountEqual(objects["foo"]["arcs"], [[0, 0], [9, 0], [0, 9], [0, 0]])

    def test_prequantize_observes_the_bounding_box(self):
        objects = {
            "foo": {"type": "LineString", "arcs": [[0, 0], [1, 0], [0, 1], [0, 0]]}
        }

        self.prequantize(objects, [-1, -1, 2, 2], 10)
        self.assertCountEqual(objects["foo"]["arcs"], [[3, 3], [6, 3], [3, 6], [3, 3]])

    def test_prequantize_applies_to_points_as_well_as_arcs(self):
        objects = {
            "foo": {
                "type": "MultiPoint",
                "coordinates": [[0, 0], [1, 0], [0, 1], [0, 0]],
            }
        }

        self.prequantize(objects, [0, 0, 1, 1], 1e4)
        self.assertCountEqual(
            objects["foo"]["coordinates"], [[0, 0], [9999, 0], [0, 9999], [0, 0]]
        )

    def test_prequantize_skips_coincident_points_in_lines(self):
        objects = {
            "foo": {
                "type": "LineString",
                "arcs": [[0, 0], [0.9, 0.9], [1.1, 1.1], [2, 2]],
            }
        }

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(objects["foo"]["arcs"], [[0, 0], [1, 1], [2, 2]])

    def test_prequantize_skips_coincident_points_in_polygons(self):
        objects = {
            "foo": {
                "type": "Polygon",
                "arcs": [[[0, 0], [0.9, 0.9], [1.1, 1.1], [2, 2], [0, 0]]],
            }
        }

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(
            objects["foo"]["arcs"], [[[0, 0], [1, 1], [2, 2], [0, 0]]]
        )

    def test_prequantize_does_not_skip_coincident_points_in_points(self):
        objects = {
            "foo": {
                "type": "MultiPoint",
                "coordinates": [[0, 0], [0.9, 0.9], [1.1, 1.1], [2, 2], [0, 0]],
            }
        }

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(
            objects["foo"]["coordinates"], [[0, 0], [1, 1], [1, 1], [2, 2], [0, 0]]
        )

    def test_prequantize_includes_closing_point_in_degenerate_lines(self):
        objects = {"foo": {"type": "LineString", "arcs": [[1, 1], [1, 1], [1, 1]]}}

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(objects["foo"]["arcs"], [[1, 1], [1, 1]])

    def test_prequantize_includes_closing_point_in_degenerate_polygons(self):
        objects = {
            "foo": {
                "type": "Polygon",
                "arcs": [[[0.9, 1], [1.1, 1], [1.01, 1], [0.9, 1]]],
            }
        }

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(
            objects["foo"]["arcs"], [[[1, 1], [1, 1], [1, 1], [1, 1]]]
        )



import unittest

from pytopojson import (
    extract,
    join,
)


class JoinTestCase(unittest.TestCase):
    def setUp(self):
        self.extract = extract.Extract()
        self.join = join.Join()

    def test_join_the_returned_hashmap_has_true_for_junction_points(self):
        junctions = self.join(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                    "ab": {"type": "LineString", "arcs": [[0, 0], [1, 0]]},
                }
            )
        )

        self.assertEqual(junctions.has([2, 0]), True)
        self.assertEqual(junctions.has([0, 0]), True)

    def test_join_the_returned_hashmap_has_undefined_for_non_junction_points(self):
        junctions = self.join(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                    "ab": {"type": "LineString", "arcs": [[0, 0], [2, 0]]},
                }
            )
        )

        self.assertEqual(junctions.has([1, 0]), False)

    def test_join_exact_duplicate_lines_abc_and_abc_have_junctions_at_their_end_points(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "abc2": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [2, 0]])

    def test_join_reversed_duplicate_lines_abc_and_cba_have_junctions_at_their_end_points(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [2, 0]])

    def test_join_exact_duplicate_rings_abca_and_abca_have_no_junctions(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                    "abca2": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_reversed_duplicate_rings_acba_and_abca_have_no_junctions(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                    "acba": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [2, 0], [1, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_rotated_duplicate_rings_bcab_and_abca_have_no_junctions(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                    "bcab": {
                        "type": "Polygon",
                        "arcs": [[[1, 0], [2, 0], [0, 0], [1, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_ring_abca_and_line_abca_have_a_junction_at_a(self):
        junctions = self.join(
            self.extract(
                {
                    "abcaLine": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [0, 0]],
                    },
                    "abcaPolygon": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0]])

    def test_join_ring_bcab_and_line_abca_have_a_junction_at_a(self):
        junctions = self.join(
            self.extract(
                {
                    "abcaLine": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [0, 0]],
                    },
                    "bcabPolygon": {
                        "type": "Polygon",
                        "arcs": [[[1, 0], [2, 0], [0, 0], [1, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0]])

    def test_join_ring_abca_and_line_bcab_have_a_junction_at_b(self):
        junctions = self.join(
            self.extract(
                {
                    "bcabLine": {
                        "type": "LineString",
                        "arcs": [[1, 0], [2, 0], [0, 0], [1, 0]],
                    },
                    "abcaPolygon": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[1, 0]])

    def test_join_when_an_old_arc_abc_extends_a_new_arc_ab_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "ab": {"type": "LineString", "arcs": [[0, 0], [1, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [1, 0], [2, 0]])

    def test_join_when_a_reversed_old_arc_cba_extends_a_new_arc_ab_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                    "ab": {"type": "LineString", "arcs": [[0, 0], [1, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [1, 0], [2, 0]])

    def test_join_when_a_new_arc_ade_shares_its_start_with_an_old_arc_abc_there_is_a_junction_at_a(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "ade": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 1], [2, 1]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [2, 0], [2, 1]])

    def test_join_ring_aba_has_no_junctions(self):
        junctions = self.join(
            self.extract(
                {"aba": {"type": "Polygon", "arcs": [[[0, 0], [1, 0], [0, 0]]]}}
            )
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_ring_aa_has_no_junctions(self):
        junctions = self.join(
            self.extract({"aa": {"type": "Polygon", "arcs": [[[0, 0], [0, 0]]]}})
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_degenerate_ring_a_has_no_junctions(self):
        junctions = self.join(
            self.extract({"a": {"type": "Polygon", "arcs": [[[0, 0]]]}})
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_when_a_new_line_dec_shares_its_end_with_an_old_line_abc_there_is_a_junction_at_c(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "dec": {"type": "LineString", "arcs": [[0, 1], [1, 1], [2, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [2, 0], [0, 1]])

    def test_join_when_a_new_line_abc_extends_an_old_line_ab_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "ab": {"type": "LineString", "arcs": [[0, 0], [1, 0]]},
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [1, 0], [2, 0]])

    def test_join_when_a_new_line_abc_extends_a_reversed_old_line_ba_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "ba": {"type": "LineString", "arcs": [[1, 0], [0, 0]]},
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [1, 0], [2, 0]])

    def test_join_when_a_new_line_starts_bc_in_the_middle_of_an_old_line_abc_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "bc": {"type": "LineString", "arcs": [[1, 0], [2, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [1, 0], [2, 0]])

    def test_join_when_a_new_line_bc_starts_in_the_middle_of_a_reversed_old_line_cba_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                    "bc": {"type": "LineString", "arcs": [[1, 0], [2, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [1, 0], [2, 0]])

    def test_join_when_a_new_line_abd_deviates_from_an_old_line_abc_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "abd": {"type": "LineString", "arcs": [[0, 0], [1, 0], [3, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [2, 0], [1, 0], [3, 0]])

    def test_join_when_a_new_line_abd_deviates_from_a_reversed_old_line_cba_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                    "abd": {"type": "LineString", "arcs": [[0, 0], [1, 0], [3, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[2, 0], [0, 0], [1, 0], [3, 0]])

    def test_join_when_a_new_line_dbc_merges_into_an_old_line_abc_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "dbc": {"type": "LineString", "arcs": [[3, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [2, 0], [1, 0], [3, 0]])

    def test_join_when_a_new_line_dbc_merges_into_a_reversed_old_line_cba_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                    "dbc": {"type": "LineString", "arcs": [[3, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[2, 0], [0, 0], [1, 0], [3, 0]])

    def test_join_when_a_new_line_dbe_shares_a_single_midpoint_with_an_old_line_abc_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "dbe": {"type": "LineString", "arcs": [[0, 1], [1, 0], [2, 1]]},
                }
            )
        )

        self.assertCountEqual(
            junctions.values(), [[0, 0], [2, 0], [2, 1], [1, 0], [0, 1]]
        )

    def test_join_when_a_new_line_abde_skips_a_point_with_an_old_line_abcde_there_is_a_junction_at_b_and_d(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abcde": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0]],
                    },
                    "abde": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [3, 0], [4, 0]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [4, 0], [1, 0], [3, 0]])

    def test_join_when_a_new_line_abde_skips_a_point_with_a_reversed_old_line_edcba_there_is_a_junction_at_b_and_d(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "edcba": {
                        "type": "LineString",
                        "arcs": [[4, 0], [3, 0], [2, 0], [1, 0], [0, 0]],
                    },
                    "abde": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [3, 0], [4, 0]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[4, 0], [0, 0], [1, 0], [3, 0]])

    def test_join_when_a_line_abcdbe_self_intersects_with_its_middle_there_are_no_junctions(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abcdbe": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]],
                    }
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [4, 0]])

    def test_join_when_a_line_abacd_self_intersects_with_its_start_there_are_no_junctions(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abacd": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [0, 0], [3, 0], [4, 0]],
                    }
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [4, 0]])

    def test_join_when_a_line_abcdbd_self_intersects_with_its_end_there_are_no_junctions(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abcdbd": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [4, 0], [3, 0], [4, 0]],
                    }
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0], [4, 0]])

    def test_join_when_an_old_line_abcdbe_self_intersects_and_shares_a_point_b_there_is_a_junction_at_b(
        self,
    ):
        junctions = self.join(
            self.extract(
                {
                    "abcdbe": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]],
                    },
                    "fbg": {"type": "LineString", "arcs": [[0, 1], [1, 0], [2, 1]]},
                }
            )
        )

        self.assertCountEqual(
            junctions.values(), [[0, 0], [4, 0], [1, 0], [0, 1], [2, 1]]
        )

    def test_join_when_a_line_abca_is_closed_there_is_a_junction_at_a(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [0, 1], [0, 0]],
                    }
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[0, 0]])

    def test_join_when_a_ring_abca_is_closed_there_are_no_junctions(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    }
                }
            )
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_exact_duplicate_rings_abca_and_abca_share_the_arc_abca(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "abca2": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_reversed_duplicate_rings_abca_and_acba_share_the_arc_abca(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "acba": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [0, 1], [1, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_coincident_rings_abca_and_bcab_share_the_arc_bcab(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "bcab": {
                        "type": "Polygon",
                        "arcs": [[[1, 0], [0, 1], [0, 0], [1, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_coincident_rings_abca_and_bacb_share_the_arc_bcab(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "bacb": {
                        "type": "Polygon",
                        "arcs": [[[1, 0], [0, 0], [0, 1], [1, 0]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [])

    def test_join_coincident_rings_abca_and_dbed_share_the_point_b(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "dbed": {
                        "type": "Polygon",
                        "arcs": [[[2, 1], [1, 0], [2, 2], [2, 1]]],
                    },
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[1, 0]])

    def test_join_coincident_ring_abca_and_line_dbe_share_the_point_b(self):
        junctions = self.join(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "dbe": {"type": "LineString", "arcs": [[2, 1], [1, 0], [2, 2]]},
                }
            )
        )

        self.assertCountEqual(junctions.values(), [[2, 1], [2, 2], [1, 0]])



import unittest

from pytopojson import geometry


class GeometryTestCase(unittest.TestCase):
    def setUp(self):
        self.geometry = geometry.Geometry()

    def test_geometry_replaces_LineString_Feature_with_LineString_Geometry(self):
        geom = self.geometry(
            {
                "foo": {
                    "type": "Feature",
                    "geometry": {"type": "LineString", "coordinates": [[0, 0]]},
                }
            }
        )

        self.assertDictEqual(geom, {"foo": {"type": "LineString", "arcs": [[0, 0]]}})

    def test_geometry_replaces_GeometryCollection_Feature_with_GeometryCollection(self):
        geom = self.geometry(
            {
                "foo": {
                    "type": "Feature",
                    "geometry": {
                        "type": "GeometryCollection",
                        "geometries": [{"type": "LineString", "coordinates": [[0, 0]]}],
                    },
                }
            }
        )

        self.assertDictEqual(
            geom,
            {
                "foo": {
                    "type": "GeometryCollection",
                    "geometries": [{"type": "LineString", "arcs": [[0, 0]]}],
                }
            },
        )

    def test_geometry_replaces_FeatureCollection_with_GeometryCollection(self):
        geom = self.geometry(
            {
                "foo": {
                    "type": "FeatureCollection",
                    "features": [
                        {
                            "type": "Feature",
                            "geometry": {"type": "LineString", "coordinates": [[0, 0]]},
                        }
                    ],
                }
            }
        )

        self.assertDictEqual(
            geom,
            {
                "foo": {
                    "type": "GeometryCollection",
                    "geometries": [{"type": "LineString", "arcs": [[0, 0]]}],
                }
            },
        )

    def test_geometry_replaces_Feature_with_null_Geometry_with_null_type_Geometry(self):
        geom = self.geometry({"foo": {"type": "Feature", "geometry": None}})

        self.assertDictEqual(geom, {"foo": {"type": None}})

    def test_geometry_replaces_top_level_null_Geometry_with_null_type_Geometry(self):
        geom = self.geometry({"foo": None})

        self.assertDictEqual(geom, {"foo": {"type": None}})

    def test_geometry_replaces_null_Geometry_in_GeometryCollection_with_null_type_Geometry(
        self,
    ):
        geom = self.geometry(
            {"foo": {"type": "GeometryCollection", "geometries": [None]}}
        )

        self.assertDictEqual(
            geom,
            {"foo": {"type": "GeometryCollection", "geometries": [{"type": None}]}},
        )

    def test_geometry_preserves_id(self):
        geom = self.geometry(
            {
                "foo": {
                    "id": "foo",
                    "type": "Feature",
                    "geometry": {"type": "LineString", "coordinates": [[0, 0]]},
                }
            }
        )

        self.assertDictEqual(
            geom, {"foo": {"id": "foo", "type": "LineString", "arcs": [[0, 0]]}}
        )

    def test_geometry_preserves_properties_if_non_empty(self):
        geom = self.geometry(
            {
                "foo": {
                    "properties": {"foo": 42},
                    "type": "Feature",
                    "geometry": {"type": "LineString", "coordinates": [[0, 0]]},
                }
            }
        )

        self.assertDictEqual(
            geom,
            {
                "foo": {
                    "properties": {"foo": 42},
                    "type": "LineString",
                    "arcs": [[0, 0]],
                }
            },
        )

    def test_geometry_applies_a_shallow_copy_for_properties(self):
        input = {
            "foo": {
                "properties": {"foo": 42},
                "type": "Feature",
                "geometry": {"type": "LineString", "coordinates": [[0, 0]]},
            }
        }

        geom = self.geometry(input)

        self.assertDictEqual(input["foo"]["properties"], geom["foo"]["properties"])

    def test_geometry_deletes_empty_properties(self):
        geom = self.geometry(
            {
                "foo": {
                    "properties": {},
                    "type": "Feature",
                    "geometry": {"type": "LineString", "coordinates": [[0, 0]]},
                }
            }
        )

        self.assertDictEqual(geom, {"foo": {"type": "LineString", "arcs": [[0, 0]]}})

    def test_geometry_does_not_convert_singular_multipoints_to_points(self):
        geom = self.geometry({"foo": {"type": "MultiPoint", "coordinates": [[0, 0]]}})

        self.assertDictEqual(
            geom, {"foo": {"type": "MultiPoint", "coordinates": [[0, 0]]}}
        )

    def test_geometry_does_not_convert_empty_multipoints_to_null(self):
        geom = self.geometry({"foo": {"type": "MultiPoint", "coordinates": []}})

        self.assertDictEqual(geom, {"foo": {"type": "MultiPoint", "coordinates": []}})

    def test_geometry_does_not_convert_singular_multilines_to_lines(self):
        geom = self.geometry(
            {"foo": {"type": "MultiLineString", "coordinates": [[[0, 0], [0, 1]]]}}
        )

        self.assertDictEqual(
            geom, {"foo": {"type": "MultiLineString", "arcs": [[[0, 0], [0, 1]]]}}
        )

    def test_geometry_does_not_convert_empty_lines_to_null(self):
        geom = self.geometry({"foo": {"type": "LineString", "coordinates": []}})

        self.assertDictEqual(geom, {"foo": {"type": "LineString", "arcs": []}})

    def test_geometry_does_not_convert_empty_multilines_to_null(self):
        geom = self.geometry(
            {
                "foo": {"type": "MultiLineString", "coordinates": []},
                "bar": {"type": "MultiLineString", "coordinates": [[]]},
            }
        )

        self.assertDictEqual(
            geom,
            {
                "foo": {"type": "MultiLineString", "arcs": []},
                "bar": {"type": "MultiLineString", "arcs": [[]]},
            },
        )

    def test_geometry_does_not_strip_empty_rings_in_polygons(self):
        geom = self.geometry(
            {
                "foo": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 0]], []],
                }
            }
        )

        self.assertDictEqual(
            geom,
            {
                "foo": {
                    "type": "Polygon",
                    "arcs": [[[0, 0], [1, 0], [1, 1], [0, 0]], []],
                }
            },
        )

    def test_geometry_does_not_strip_empty_lines_in_multilines(self):
        geom = self.geometry(
            {
                "foo": {
                    "type": "MultiLineString",
                    "coordinates": [
                        [[0, 0], [1, 0], [1, 1], [0, 0]],
                        [],
                        [[0, 0], [1, 0]],
                    ],
                }
            }
        )

        self.assertDictEqual(
            geom,
            {
                "foo": {
                    "type": "MultiLineString",
                    "arcs": [[[0, 0], [1, 0], [1, 1], [0, 0]], [], [[0, 0], [1, 0]]],
                }
            },
        )

    def test_geometry_does_not_convert_empty_polygons_to_null(self):
        geom = self.geometry(
            {
                "foo": {"type": "Polygon", "coordinates": []},
                "bar": {"type": "Polygon", "coordinates": [[]]},
            }
        )

        self.assertDictEqual(
            geom,
            {
                "foo": {"type": "Polygon", "arcs": []},
                "bar": {"type": "Polygon", "arcs": [[]]},
            },
        )

    def test_geometry_does_not_strip_empty_polygons_in_multipolygons(self):
        geom = self.geometry(
            {
                "foo": {
                    "type": "MultiPolygon",
                    "coordinates": [[[[0, 0], [1, 0], [1, 1], [0, 0]], []], [], [[]]],
                }
            }
        )

        self.assertDictEqual(
            geom,
            {
                "foo": {
                    "type": "MultiPolygon",
                    "arcs": [[[[0, 0], [1, 0], [1, 1], [0, 0]], []], [], [[]]],
                }
            },
        )

    def test_geometry_does_not_convert_singular_multipolygons_to_polygons(self):
        geom = self.geometry(
            {
                "foo": {
                    "type": "MultiPolygon",
                    "coordinates": [[[[0, 0], [0, 1], [1, 0], [0, 0]]]],
                }
            }
        )

        self.assertDictEqual(
            geom,
            {
                "foo": {
                    "type": "MultiPolygon",
                    "arcs": [[[[0, 0], [0, 1], [1, 0], [0, 0]]]],
                }
            },
        )



import json
import math
import unittest

from pytopojson import bbox


class BBoxTestCase(unittest.TestCase):
    def setUp(self):
        self.bbox = bbox.BBox()

    @staticmethod
    def _load_json_file(filepath):
        with open(filepath, "r") as src:
            return json.load(src)

    def test_bbox_topology_ignores_the_existing_bbox_if_any(self):
        bbox = [1, 2, 3, 4]
        self.assertListEqual(
            self.bbox({"type": "Topology", "bbox": bbox, "objects": {}, "arcs": {}}),
            [math.inf, math.inf, -math.inf, -math.inf],
        )

    def test_bbox_topology_computes_the_bbox_for_a_quantized_topology_if_missing(self):
        topology = self._load_json_file("tests/client/topojson/polygon-q1e4.json")
        self.assertListEqual(self.bbox(topology), [0, 0, 10, 10])

    def test_bbox_topology_computes_the_bbox_for_a_non_quantized_topology_if_missing(
        self,
    ):
        topology = self._load_json_file("tests/client/topojson/polygon.json")
        self.assertListEqual(self.bbox(topology), [0, 0, 10, 10])

    def test_bbox_topology_considers_points(self):
        topology = self._load_json_file("tests/client/topojson/point.json")
        self.assertListEqual(self.bbox(topology), [0, 0, 10, 10])

    def test_bbox_topology_considers_multipoints(self):
        topology = self._load_json_file("tests/client/topojson/points.json")
        self.assertListEqual(self.bbox(topology), [0, 0, 10, 10])



from pytopojson.join import Join


class Cut(object):
    """
    Given an extracted (pre-)topology, cuts (or rotates) arcs so that all shared
    point sequences are identified. The topology can then be subsequently deduped
    to remove exact duplicate arcs.
    """

    def __init__(self):
        self.join = Join()

    def __call__(self, topology, *args, **kwargs):
        self.junctions = self.join(topology)
        self.coordinates = topology["coordinates"]
        self.lines = list()
        self.rings = list()
        self.index = 0
        self.cut_geometry_type = {
            "GeometryCollection": self._geometry_collection_call,
            "LineString": self._line_string_call,
            "MultiLineString": self._multi_line_string_call,
            "Polygon": self._polygon_call,
            "MultiPolygon": self._multi_polygon_call,
        }

        for k, v in topology["objects"].items():
            self.cut_geometry(v)

        topology["lines"] = self.lines
        topology["rings"] = self.rings
        return topology

    def rotate_array(self, array, start, end, offset):
        self.reverse(array, start, end)
        self.reverse(array, start, start + offset)
        self.reverse(array, start + offset, end)

    def reverse(self, array, start, end):
        mid = start + ((end - start) >> 1)
        end -= 1

        while start < mid:
            t = array[start]
            array[start] = array[end]
            array[end] = t

            start += 1
            end -= 1

    def _line_string_call(self, o):
        o["arcs"] = self.cut_line(o["arcs"])

    def _multi_line_string_call(self, o):
        o["arcs"] = list(map(self.cut_line, o["arcs"]))

    def _polygon_call(self, o):
        o["arcs"] = list(map(self.cut_ring, o["arcs"]))

    def _multi_polygon_call(self, o):
        o["arcs"] = list(map(self.cut_multi_ring, o["arcs"]))

    def _geometry_collection_call(self, o):
        for geometry in o["geometries"]:
            self.cut_geometry(geometry)

    def cut_line(self, line):
        line = dict({0: line[0], 1: line[1]})
        line_mid = line[0] + 1
        line_end = line[1]

        l = line
        while line_mid < line_end:
            if self.junctions.has(self.coordinates[line_mid]):
                next = {0: line_mid, 1: l[1]}
                l[1] = line_mid
                l["next"] = next
                l = next

            line_mid += 1

        self.lines.append(line)
        return line

    def cut_ring(self, ring):
        ring = dict({0: ring[0], 1: ring[1]})
        ring_start = ring[0]
        ring_mid = ring_start
        ring_end = ring[1]
        ring_fixed = self.junctions.has(self.coordinates[ring_mid])

        ring_mid += 1
        r = ring
        while ring_mid < ring_end:
            if self.junctions.has(self.coordinates[ring_mid]):
                if ring_fixed:
                    next = {0: ring_mid, 1: r[1]}
                    r[1] = ring_mid
                    r["next"] = next
                    r = next
                # For the first junction, we can rotate rather than cut.
                else:
                    self.rotate_array(
                        self.coordinates, ring_start, ring_end, ring_end - ring_mid
                    )
                    self.coordinates[ring_end] = self.coordinates[ring_start]
                    ring_fixed = True
                    ring_mid = ring_start

            ring_mid += 1

        self.rings.append(ring)
        return ring

    def cut_multi_ring(self, rings):
        return list(map(self.cut_ring, rings))

    def cut_geometry(self, geometry):
        if geometry and geometry["type"] in self.cut_geometry_type:
            self.cut_geometry_type[geometry["type"]](geometry)



import argparse
import json
import os
import sys

from pytopojson import topology


__version__ = "0.0.2"


def read(specifiers):
    objs = dict()

    for specifier in specifiers:
        i = specifier.find("=")
        if 0 <= i:
            name, file = specifier.split("=")
        else:
            file = specifier
            filename = os.path.basename(file)
            name, _ = os.path.splitext(filename)

        if name in objs:
            print(f"\n  error: object {name} is not unique\n")

        else:
            objs[name] = read_object(file)

    return objs


def read_object(filepath):
    with open(filepath, "r") as f:
        data = f.read()
    return json.loads(data)


def read_newline_delimited_object():
    pass


def write(data, output):
    if isinstance(output, str):
        with open(output, "w") as dst:
            json.dump(data, dst, separators=(",", ":"))
            dst.write("\n")
    else:
        json.dump(data, output, separators=(",", ":"))
        sys.stdout.write("\n")


def main():
    # Create OptionParser object and set options
    parser = argparse.ArgumentParser(
        description="Converts GeoJSON features to TopoJSON objects."
    )

    parser.add_argument(
        "-o",
        "--out",
        dest="file",
        help="output file name; defaults to “-” for stdout",
        default="-",
    )
    # parser.add_argument('-n', '--newline-delimited',
    #                     help='accept newline-delimited JSON')
    parser.add_argument(
        "-q",
        "--quantization",
        dest="count",
        type=float,
        help="pre-quantization parameter; 0 disables quantization",
        default=0,
    )
    parser.add_argument(
        "geojsons",
        metavar="[name=]file",
        type=str,
        nargs="+",
        help="file with GeoJSON data",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s v{__version__}"
    )

    # Parse and store the command-line arguments in dictionary
    opts = parser.parse_args()
    kwargs = vars(opts)

    # Output destination
    out = kwargs["file"] if kwargs["file"] != "-" else sys.stdout

    # Quantization
    quant = {"quantization": kwargs["count"]}
    quant = quant if quant["quantization"] != 0 else dict()

    # Read files and compute topology
    objects = read(kwargs["geojsons"])
    topology_ = topology.Topology()
    topo = topology_(objects, **quant)

    # Write TopoJSON
    write(topo, out)


if __name__ == "__main__":
    main()



import unittest

from pytopojson import extract


class ExtractTestCase(unittest.TestCase):
    def setUp(self):
        self.extract = extract.Extract()

    def test_extract_copies_coordinates_sequentially_into_a_buffer(self):
        topology = self.extract(
            {
                "foo": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
            }
        )

        self.assertCountEqual(
            topology["coordinates"], [[0, 0], [1, 0], [2, 0], [0, 0], [1, 0], [2, 0]]
        )

    def test_extract_does_not_copy_point_geometries_into_the_coordinate_buffer(self):
        topology = self.extract(
            {
                "foo": {"type": "Point", "arcs": [0, 0]},
                "bar": {"type": "MultiPoint", "arcs": [[0, 0], [1, 0], [2, 0]]},
            }
        )

        self.assertCountEqual(topology["coordinates"], [])
        self.assertCountEqual(topology["objects"]["foo"]["arcs"], [0, 0])
        self.assertCountEqual(
            topology["objects"]["bar"]["arcs"], [[0, 0], [1, 0], [2, 0]]
        )

    def test_extract_includes_closing_coordinates_in_polygons(self):
        topology = self.extract(
            {"foo": {"type": "Polygon", "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]]}}
        )

        self.assertCountEqual(topology["coordinates"], [[0, 0], [1, 0], [2, 0], [0, 0]])

    def test_extract_represents_lines_as_contiguous_slices_of_the_coordinate_buffer(
        self,
    ):
        topology = self.extract(
            {
                "foo": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
            }
        )

        self.assertDictEqual(
            topology["objects"],
            {
                "foo": {"type": "LineString", "arcs": [0, 2]},
                "bar": {"type": "LineString", "arcs": [3, 5]},
            },
        )

    def test_extract_represents_rings_as_contiguous_slices_of_the_coordinate_buffer(
        self,
    ):
        topology = self.extract(
            {
                "foo": {"type": "Polygon", "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]]},
                "bar": {"type": "Polygon", "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]]},
            }
        )

        self.assertDictEqual(
            topology["objects"],
            {
                "foo": {"type": "Polygon", "arcs": [[0, 3]]},
                "bar": {"type": "Polygon", "arcs": [[4, 7]]},
            },
        )

    def test_extract_exposes_the_constructed_lines_and_rings_in_the_order_of_construction(
        self,
    ):
        topology = self.extract(
            {
                "line": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                "multiline": {
                    "type": "MultiLineString",
                    "arcs": [[[0, 0], [1, 0], [2, 0]]],
                },
                "polygon": {
                    "type": "Polygon",
                    "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                },
            }
        )

        self.assertCountEqual(topology["lines"], [[0, 2], [3, 5]])
        self.assertCountEqual(topology["rings"], [[6, 9]])

    def test_extract_supports_nested_geometry_collections(self):
        topology = self.extract(
            {
                "foo": {
                    "type": "GeometryCollection",
                    "geometries": [
                        {
                            "type": "GeometryCollection",
                            "geometries": [
                                {"type": "LineString", "arcs": [[0, 0], [0, 1]]}
                            ],
                        }
                    ],
                }
            }
        )

        self.assertDictEqual(
            topology["objects"]["foo"],
            {
                "type": "GeometryCollection",
                "geometries": [
                    {
                        "type": "GeometryCollection",
                        "geometries": [{"type": "LineString", "arcs": [0, 1]}],
                    }
                ],
            },
        )



import unittest

from pytopojson import untransform


class UntransformTestCase(unittest.TestCase):
    def setUp(self):
        self.untransform = untransform.Untransform()

    def test_untransform_topology_returns_the_identity_function_if_transform_is_undefined(
        self,
    ):
        untransform = self.untransform(None)
        point = dict()
        self.assertDictEqual(untransform(point), point)

    def test_untransform_topology_returns_a_point_transform_function_if_transform_is_defined(
        self,
    ):
        untransform = self.untransform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(untransform([16, 26]), [6, 7])

    def test_untransform_point_returns_a_new_point(self):
        untransform = self.untransform({"scale": [2, 3], "translate": [4, 5]})
        point = [16, 26]
        self.assertListEqual(untransform(point), [6, 7])
        self.assertListEqual(point, [16, 26])

    def test_untransform_point_preserves_extra_dimensions(self):
        untransform = self.untransform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(untransform([16, 26, 42]), [6, 7, 42])
        self.assertListEqual(untransform([16, 26, "foo"]), [6, 7, "foo"])
        self.assertListEqual(untransform([16, 26, "foo", 42]), [6, 7, "foo", 42])

    def test_untransform_point_untransforms_individual_points(self):
        untransform = self.untransform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(untransform([6, 11]), [1, 2])
        self.assertListEqual(untransform([10, 17]), [3, 4])
        self.assertListEqual(untransform([14, 23]), [5, 6])

    def test_untransform_point_index_untransforms_delta_encoded_arcs(self):
        untransform = self.untransform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(untransform([6, 11], 0), [1, 2])
        self.assertListEqual(untransform([12, 23], 1), [3, 4])
        self.assertListEqual(untransform([22, 41], 2), [5, 6])
        self.assertListEqual(untransform([24, 47], 3), [1, 2])
        self.assertListEqual(untransform([30, 59], 4), [3, 4])
        self.assertListEqual(untransform([40, 77], 5), [5, 6])

    def test_untransform_point_index_untransforms_multiple_delta_encoded_arcs(self):
        untransform = self.untransform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(untransform([6, 11], 0), [1, 2])
        self.assertListEqual(untransform([12, 23], 1), [3, 4])
        self.assertListEqual(untransform([22, 41], 2), [5, 6])
        self.assertListEqual(untransform([6, 11], 0), [1, 2])
        self.assertListEqual(untransform([12, 23], 1), [3, 4])
        self.assertListEqual(untransform([22, 41], 2), [5, 6])



import math

from pytopojson import transform


class BBox(object):
    def __init__(self):
        self.transform = transform.Transform()
        self.x_0 = math.inf
        self.y_0 = self.x_0
        self.x_1 = -self.x_0
        self.y_1 = -self.x_0
        self.t = None

    def __call__(self, topology, *args, **kwargs):
        self.t = self.transform(topology.get("transform", None))

        for arc in topology["arcs"]:
            i = 0
            n = len(arc)

            while i < n:
                p = self.t(arc[i], i)
                if p[0] < self.x_0:
                    self.x_0 = p[0]
                if self.x_1 < p[0]:
                    self.x_1 = p[0]
                if p[1] < self.y_0:
                    self.y_0 = p[1]
                if self.y_1 < p[1]:
                    self.y_1 = p[1]
                i += 1

        for k in topology["objects"]:
            self.bbox_geometry(topology["objects"][k])

        return [self.x_0, self.y_0, self.x_1, self.y_1]

    def bbox_point(self, p):
        p = self.t(p)
        if p[0] < self.x_0:
            self.x_0 = p[0]
        if self.x_1 < p[0]:
            self.x_1 = p[0]
        if p[1] < self.y_0:
            self.y_0 = p[1]
        if self.y_1 < p[1]:
            self.y_1 = p[1]

    def bbox_geometry(self, o):
        if o["type"] == "GeometryCollection":
            for geom in o["geometries"]:
                self.bbox_geometry(geom)
        elif o["type"] == "Point":
            self.bbox_point(o["coordinates"])
        elif o["type"] == "MultiPoint":
            for coord in o["coordinates"]:
                self.bbox_point(coord)



import math

from pytopojson.commons import Array


class HashMap(object):
    def __init__(
        self, size, hash, equal, key_type=Array, key_empty=None, value_type=Array
    ):
        self.hash = hash
        self.equal = equal
        self.key_empty = key_empty
        self.size = 1 << max(4, int(math.ceil(math.log(size + 1e-9) / math.log(2))))
        self.key_store = key_type(self.size)
        self.val_store = value_type(self.size)
        self.mask = self.size - 1

        for i in range(self.size):
            self.key_store[i] = key_empty

        self.value = {
            "set": self.set,
            "maybeSet": self.maybe_set,
            "get": self.get,
            "keys": self.keys,
        }

    def set(self, key, value):
        index = self.hash(key) & self.mask
        match_key = self.key_store[index]
        collisions = 0

        while match_key != self.key_empty:
            if self.equal(match_key, key):
                self.val_store[index] = value
                return self.val_store[index]

            collisions += 1

            if self.size <= collisions:
                raise ValueError("Full HashMap")

            index = (index + 1) & self.mask
            match_key = self.key_store[index]

        self.key_store[index] = key
        self.val_store[index] = value
        return value

    def maybe_set(self, key, value):
        index = self.hash(key) & self.mask
        match_key = self.key_store[index]
        collisions = 1

        while match_key != self.key_empty:
            if self.equal(match_key, key):
                return self.val_store[index]

            collisions += 1
            if self.size <= collisions:
                raise ValueError("Full HashMap")

            index = (index + 1) & self.mask
            match_key = self.key_store[index]

        self.key_store[index] = key
        self.val_store[index] = value
        return value

    def get(self, key, missing_value=None):
        index = self.hash(key) & self.mask
        match_key = self.key_store[index]
        collisions = 1

        while match_key != self.key_empty:
            if self.equal(match_key, key):
                return self.val_store[index]
            if self.size <= collisions:
                break

            collisions += 1
            index = (index + 1) & self.mask
            match_key = self.key_store[index]

        return missing_value

    def keys(self):
        keys = list()
        i = 0
        n = len(self.key_store)

        # TODO: Check for k in self.key_store
        while i < n:
            match_key = self.key_store[i]
            if match_key != self.key_empty:
                keys.append(match_key)

        return keys


class HashSet(object):
    def __init__(self, size, hash, equal, type=Array, empty=None):
        self.size = 1 << max(4, int(math.ceil(math.log(size + 1e-9) / math.log(2))))
        self.hash = hash
        self.equal = equal
        self.empty = empty
        self.store = type(self.size)
        self.mask = self.size - 1

        for i in range(self.size):
            self.store[i] = empty

        self.value = {"add": self.add, "has": self.has, "values": self.values}

    def add(self, value):
        index = self.hash(value) & self.mask
        match = self.store[index]
        collisions = 0

        while match != self.empty:
            if self.equal(match, value):
                return True

            collisions += 1
            if self.size <= collisions:
                raise ValueError("Full hashset")

            index = (index + 1) & self.mask
            match = self.store[index]

        self.store[index] = value
        return True

    def has(self, value):
        index = self.hash(value) & self.mask
        match = self.store[index]
        collisions = 0

        while match != self.empty:
            if self.equal(match, value):
                return True

            collisions += 1
            if self.size <= collisions:
                break

            index = (index + 1) & self.mask
            match = self.store[index]

        return False

    def values(self):
        values = list()
        n = len(self.store)
        i = 0
        while i < n:
            match = self.store[i]
            if match != self.empty:
                values.append(match)

            i += 1

        return values



from pytopojson.hash.hash import HashMap, HashSet
from pytopojson.hash.point import equal as equal_point
from pytopojson.hash.point import hash as hash_point
from pytopojson.commons import Int8Array, Int32Array


# Computes the bounding box of the specified hash of GeoJSON objects.
class Join(object):
    """
    Given an extracted (pre-)topology, identifies all of the junctions. These are
    the points at which arcs (lines or rings) will need to be cut so that each
    arc is represented uniquely.

    A junction is a point where at least one arc deviates from another arc going
    through the same point. For example, consider the point B. If there is a arc
    through ABC and another arc through CBA, then B is not a junction because in
    both cases the adjacent point pairs are {A,C}. However, if there is an
    additional arc ABD, then {A,D} != {A,C}, and thus B becomes a junction.

    For a closed ring ABCA, the first point A's adjacent points are the second
    and last point {B,C}. For a line, the first and last point are always
    considered junctions, even if the line is closed; this ensures that a closed
    line is never rotated.
    """

    def __init__(self):
        pass

    def __call__(self, topology, *args, **kwargs):
        self.coordinates = topology["coordinates"]
        self.lines = topology["lines"]
        self.rings = topology["rings"]
        self.indexes = self.index()
        self.visited_by_index = Int32Array(len(self.coordinates))
        self.left_by_index = Int32Array(len(self.coordinates))
        self.right_by_index = Int32Array(len(self.coordinates))
        self.junction_by_index = Int8Array(len(self.coordinates))
        self.junction_count = 0

        for i in range(len(self.coordinates)):
            self.visited_by_index[i] = self.left_by_index[i] = self.right_by_index[
                i
            ] = -1

        for i in range(len(self.lines)):
            line = self.lines[i]
            line_start = line[0]
            line_end = line[1]

            current_index = self.indexes[line_start]
            line_start += 1
            next_index = self.indexes[line_start]
            self.junction_count += 1
            self.junction_by_index[current_index] = 1
            line_start += 1

            while line_start <= line_end:
                previous_index = current_index
                current_index = next_index
                next_index = self.indexes[line_start]
                self.sequence(i, previous_index, current_index, next_index)
                line_start += 1

            self.junction_count += 1
            self.junction_by_index[next_index] = 1

        for i in range(len(self.coordinates)):
            self.visited_by_index[i] = -1

        for i in range(len(self.rings)):
            ring = self.rings[i]
            ring_start = ring[0] + 1
            ring_end = ring[1]

            previous_index = self.indexes[ring_end - 1]
            current_index = self.indexes[ring_start - 1]
            next_index = self.indexes[ring_start]

            self.sequence(i, previous_index, current_index, next_index)

            ring_start += 1

            while ring_start <= ring_end:
                previous_index = current_index
                current_index = next_index
                next_index = self.indexes[ring_start]
                self.sequence(i, previous_index, current_index, next_index)
                ring_start += 1

        self.visited_by_index = self.left_by_index = self.right_by_index = None
        self.junction_by_point = HashSet(
            self.junction_count * 1.4, hash_point, equal_point
        )

        len_junction_by_index = len(self.junction_by_index)
        for i in range(len(self.coordinates)):
            j = self.indexes[i]
            if j < len_junction_by_index and self.junction_by_index[j]:
                self.junction_by_point.add(self.coordinates[j])

        return self.junction_by_point

    def sequence(self, i, previous_index, current_index, next_index):
        # ignore self-intersection
        if self.visited_by_index[current_index] == i:
            return

        self.visited_by_index[current_index] = i
        left_index = self.left_by_index[current_index]

        if 0 <= left_index:
            right_index = self.right_by_index[current_index]
            if (left_index != previous_index or right_index != next_index) and (
                left_index != next_index or right_index != previous_index
            ):
                self.junction_count += 1
                self.junction_by_index[current_index] = 1

        else:
            self.left_by_index[current_index] = (
                0 if previous_index is None else previous_index
            )
            self.right_by_index[current_index] = 0 if next_index is None else next_index

    def index(self):
        index_by_point = HashMap(
            len(self.coordinates) * 1.4,
            self.hash_index,
            self.equal_index,
            Int32Array,
            -1,
            Int32Array,
        )
        indexes = Int32Array(len(self.coordinates))

        for i in range(len(self.coordinates)):
            tmp = index_by_point.maybe_set(i, i)
            indexes[i] = tmp

        return indexes

    def hash_index(self, i):
        return hash_point(self.coordinates[i])

    def equal_index(self, i, j):
        return equal_point(self.coordinates[i], self.coordinates[j])



import argparse
import json
import os
import sys

from pytopojson import feature


__version__ = "0.0.2"


def read(input_):
    if isinstance(input_, str):
        with open(input_, "r") as src:
            data = json.load(src)
    else:
        data = json.load(input_)

    return data


def write_list(topology_):
    for name in topology_["objects"]:
        print(name)


def read_newline_delimited_object():
    pass


def write(topology_, objects):
    feat = feature.Feature()
    for specifier in objects:
        i = specifier.find("=")
        if 0 <= i:
            name, file_ = specifier.split("=")
        else:
            file_ = specifier
            filename = os.path.basename(file_)
            name, _ = os.path.splitext(filename)

        if name not in topology_["objects"]:
            print(f"\n  error: object {name} not found\n")
            return

        write_feature(file_, feat(topology, name))


def write_feature(output, feat):
    if output == "-":
        json.dump(feat, sys.stdout, separators=(",", ":"))
        sys.stdout.write("\n")
    else:
        with open(output, "w") as dst:
            json.dump(feat, dst, separators=(",", ":"))
            dst.write("\n")


def main():
    # Create OptionParser object and set options
    parser = argparse.ArgumentParser(
        description="Converts TopoJSON objects to GeoJSON features."
    )

    parser.add_argument(
        "-i",
        "--in",
        dest="file",
        help='input topology file name; defaults to "-" for stdin',
        default="-",
    )
    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        default=False,
        help="list the object names on the input topology",
    )
    parser.add_argument(
        "objects",
        metavar="[geometry_type=]file",
        type=str,
        nargs="+",
        help="output with GeoJSON data for given geometry in input",
    )
    # parser.add_argument('-n', '--newline-delimited',
    #                     help='output newline-delimited JSON')
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s v{__version__}"
    )

    # Parse and store the command-line arguments in dictionary
    opts = parser.parse_args()
    kwargs = vars(opts)

    # Read input file
    input_ = kwargs["file"] if kwargs["file"] != "-" else sys.stdin
    topology = read(input_)

    if kwargs["list"]:
        write_list(topology)
    else:
        write(topology, kwargs["objects"])


if __name__ == "__main__":
    main()



import unittest

from pytopojson import delta


class DeltaTestCase(unittest.TestCase):
    def setUp(self):
        self.delta = delta.Delta()

    def test_delta_converts_arcs_to_delta_encoding(self):
        d = self.delta([[[0, 0], [9999, 0], [0, 9999], [0, 0]]])

        self.assertCountEqual(d, [[[0, 0], [9999, 0], [-9999, 9999], [0, -9999]]])

    def test_delta_skips_coincident_points(self):
        d = self.delta([[[0, 0], [9999, 0], [9999, 0], [0, 9999], [0, 0]]])

        self.assertCountEqual(d, [[[0, 0], [9999, 0], [-9999, 9999], [0, -9999]]])

    def test_delta_preserves_at_least_two_positions(self):
        d = self.delta(
            [[[12345, 12345], [12345, 12345], [12345, 12345], [12345, 12345]]]
        )

        self.assertCountEqual(d, [[[12345, 12345], [0, 0]]])



import argparse
import json
import os
import sys

from pytopojson import quantize


__version__ = "0.0.2"


def read(input_):
    if isinstance(input_, str):
        with open(input_, "r") as src:
            return json.load(src)
    else:
        return json.load(sys.stdin)


def quantize_topology(topology, quantization):
    quantize_ = quantize.Quantize()
    return quantize_(topology, quantization)


def write(output, topology):
    if isinstance(output, str):
        with open(output, "w") as dst:
            json.dump(topology, dst, separators=(",", ":"))
            dst.write("\n")
    else:
        json.dump(topology, sys.stdout, separators=(",", ":"))
        sys.stdout.write("\n")


def _valid_quantization_parameter(arg):
    try:
        f = float(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be a floating point number")
    if f <= 2.0:
        raise argparse.ArgumentTypeError(f"Invalid quantization parameter {arg}")
    return f


def main():
    # Create OptionParser object and set options
    parser = argparse.ArgumentParser(description="Quantizes TopoJSON.")

    parser.add_argument(
        "-i",
        "--in",
        dest="input",
        default="-",
        help='input topology file name; defaults to "-" for stdin',
    )
    parser.add_argument(
        "-o",
        "--out",
        dest="output",
        default="-",
        help='output topology file name; defaults to "-" for stdout',
    )
    parser.add_argument(
        "-q",
        "--quantization",
        dest="quantization",
        type=_valid_quantization_parameter,
        required=True,
        help="quantization value",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s v{__version__}"
    )

    # Parse and store the command-line arguments in dictionary
    opts = parser.parse_args()
    kwargs = vars(opts)

    # Read input file
    input_ = kwargs["input"] if kwargs["input"] != "-" else sys.stdin
    output = kwargs["output"] if kwargs["output"] != "-" else sys.stdout

    topology = read(input_)
    quantized = quantize_topology(topology, kwargs["quantization"])
    write(output, quantized)


if __name__ == "__main__":
    main()


import json
import subprocess
import unittest

from tests import in_delta


class Topo2GeoTestCase(unittest.TestCase):
    def topo2geo(self, command, result):
        command = " ".join(command.split()).split(" ")
        p = subprocess.run(["python", "bin/topo2geo.py", *command], capture_output=True)
        out, err = p.stdout, p.stderr
        actual = json.loads(out)
        with open(result, "r") as f:
            expected = json.load(f)

        self.assertTrue(in_delta.in_delta(actual, expected))

    def test_polygon(self):
        # self.topo2geo('polygon=- < tests/client/topojson/polygon.json',
        self.topo2geo(
            "polygon=- -i tests/client/topojson/polygon.json",
            "tests/client/geojson/polygon.json",
        )

    def test_polygon_quantized(self):
        # self.topo2geo('polygon=- < tests/client/topojson/polygon-q1e4.json',
        self.topo2geo(
            "polygon=- -i tests/client/topojson/polygon-q1e4.json",
            "tests/client/geojson/polygon.json",
        )

    def test_polygon_quantized_2(self):
        # self.topo2geo('polygon=- < tests/client/topojson/polygon-q1e5.json',
        self.topo2geo(
            "polygon=- -i tests/client/topojson/polygon-q1e5.json",
            "tests/client/geojson/polygon.json",
        )

    def test_polygon_mercator(self):
        # self.topo2geo('polygon=- < tests/client/topojson/polygon-mercator.json',
        self.topo2geo(
            "polygon=- -i tests/client/topojson/polygon-mercator.json",
            "tests/client/geojson/polygon-mercator.json",
        )

    def test_polygon_mercator_quantized(self):
        # self.topo2geo('polygon=- < tests/client/topojson/polygon-mercator-q1e4.json',
        self.topo2geo(
            "polygon=- -i tests/client/topojson/polygon-mercator-q1e4.json",
            "tests/client/geojson/polygon-mercator.json",
        )

    def test_polygon_mercator_quantized_2(self):
        # self.topo2geo('polygon=- < tests/client/topojson/polygon-mercator-q1e5.json',
        self.topo2geo(
            "polygon=- -i tests/client/topojson/polygon-mercator-q1e5.json",
            "tests/client/geojson/polygon-mercator.json",
        )

    # def test_polygon_united_states(self):
    #     # self.topo2geo('polygon=- < tests/client/topojson/polygon-mercator-q1e5.json',
    #     self.topo2geo(
    #         "us-states=- -i tests/client/topojson/us-states.json",
    #         "tests/client/geojson/us-states.json",
    #     )



from pytopojson import identity


class Transform(object):
    def __init__(self):
        self.identity = identity.Identity()
        self.x_0, self.y_0 = 0, 0
        self.k_x, self.k_y = 0, 0
        self.d_x, self.d_y = 0, 0

    def __call__(self, transform=None, *args, **kwargs):
        if transform is None:
            return self.identity

        self.k_x, self.k_y = transform["scale"]
        self.d_x, self.d_y = transform["translate"]

        return self.func

    def func(self, input, i=None):
        if i is None or i == 0:
            self.x_0, self.y_0 = 0, 0

        output = input.copy()

        self.x_0 += input[0]
        self.y_0 += input[1]
        output[0] = self.x_0 * self.k_x + self.d_x
        output[1] = self.y_0 * self.k_y + self.d_y

        return output



import json
import math
import unittest

from pytopojson import neighbors


class NeighborsTestCase(unittest.TestCase):
    def setUp(self):
        self.neighbors = neighbors.Neighbors()

    def test_neighbors_returns_an_empty_array_for_objects_with_no_neighbors(self):
        """
        A-----B

        C-----D
        """
        topology = {
            "type": "Topology",
            "objects": {
                "ab": {"type": "LineString", "arcs": [0]},
                "cd": {"type": "LineString", "arcs": [1]},
            },
            "arcs": [[[0, 0], [1, 0]], [[0, 1], [1, 1]]],
        }
        self.assertListEqual(
            self.neighbors([topology["objects"]["ab"], topology["objects"]["cd"]]),
            [[], []],
        )

    def test_neighbors_returns_an_empty_array_for_empty_input(self):
        self.assertListEqual(self.neighbors([]), [])

    def test_neighbors_geometries_that_only_share_isolated_points_are_not_considered_neighbors(
        self,
    ):
        """
        A-----B-----C
        """
        topology = {
            "type": "Topology",
            "objects": {
                "ab": {"type": "LineString", "arcs": [0]},
                "bc": {"type": "LineString", "arcs": [1]},
            },
            "arcs": [[[0, 0], [1, 0]], [[1, 0], [2, 0]]],
        }
        self.assertListEqual(
            self.neighbors([topology["objects"]["ab"], topology["objects"]["bc"]]),
            [[], []],
        )

    def test_neighbors_geometries_that_share_arcs_are_considered_neighbors(self):
        """
        A-----B-----C-----D
        """
        topology = {
            "type": "Topology",
            "objects": {
                "abc": {"type": "LineString", "arcs": [0, 1]},
                "bcd": {"type": "LineString", "arcs": [1, 2]},
            },
            "arcs": [[[0, 0], [1, 0]], [[1, 0], [2, 0]], [[2, 0], [3, 0]]],
        }
        self.assertListEqual(
            self.neighbors([topology["objects"]["abc"], topology["objects"]["bcd"]]),
            [[1], [0]],
        )

    def test_neighbors_geometries_that_share_reversed_arcs_are_considered_neighbors(
        self,
    ):
        """
        A-----B-----C-----D
        """
        topology = {
            "type": "Topology",
            "objects": {
                "abc": {"type": "LineString", "arcs": [0, 1]},
                "dcb": {"type": "LineString", "arcs": [2, -2]},
            },
            "arcs": [[[0, 0], [1, 0]], [[1, 0], [2, 0]], [[3, 0], [2, 0]]],
        }
        self.assertListEqual(
            self.neighbors([topology["objects"]["abc"], topology["objects"]["dcb"]]),
            [[1], [0]],
        )

    def test_neighbors_neighbors_are_returned_in_sorted_order_by_index(self):
        """
        A-----B-----C-----D-----E-----F
        """
        topology = {
            "type": "Topology",
            "objects": {
                "abcd": {"type": "LineString", "arcs": [0, 1, 2]},
                "bcde": {"type": "LineString", "arcs": [1, 2, 3]},
                "cdef": {"type": "LineString", "arcs": [2, 3, 4]},
                "dbca": {"type": "LineString", "arcs": [-3, -2, -1]},
                "edcb": {"type": "LineString", "arcs": [-4, -3, -2]},
                "fedc": {"type": "LineString", "arcs": [-5, -4, -3]},
            },
            "arcs": [
                [[0, 0], [1, 0]],
                [[1, 0], [2, 0]],
                [[2, 0], [3, 0]],
                [[3, 0], [4, 0]],
                [[4, 0], [5, 0]],
            ],
        }
        self.assertListEqual(
            self.neighbors(
                [
                    topology["objects"]["abcd"],
                    topology["objects"]["bcde"],
                    topology["objects"]["cdef"],
                    topology["objects"]["dbca"],
                    topology["objects"]["edcb"],
                    topology["objects"]["fedc"],
                ]
            ),
            [
                [1, 2, 3, 4, 5],
                [0, 2, 3, 4, 5],
                [0, 1, 3, 4, 5],
                [0, 1, 2, 4, 5],
                [0, 1, 2, 3, 5],
                [0, 1, 2, 3, 4],
            ],
        )

    def test_neighbors_the_polygons_abcda_and_befcb_are_neighbors_but_ghig_is_not(self):
        """
         A-----B-----E     G
         |     |     |     |\
         |     |     |     | \
         |     |     |     |  \
         |     |     |     |   \
         |     |     |     |    \
         D-----C-----F     I-----H
        """
        topology = {
            "type": "Topology",
            "objects": {
                "abcda": {"type": "Polygon", "arcs": [[0, 1]]},
                "befcb": {"type": "Polygon", "arcs": [[2, -1]]},
                "ghig": {"type": "Polygon", "arcs": [[3]]},
            },
            "arcs": [
                [[1, 0], [1, 1]],
                [[1, 1], [0, 1], [0, 0], [1, 0]],
                [[1, 0], [2, 0], [2, 1], [1, 1]],
                [[3, 0], [4, 1], [3, 1], [3, 0]],
            ],
        }
        self.assertListEqual(
            self.neighbors(
                [
                    topology["objects"]["abcda"],
                    topology["objects"]["befcb"],
                    topology["objects"]["ghig"],
                ]
            ),
            [[1], [0], []],
        )

    def test_neighbors_the_polygons_abedghkja_and_bclkhifeb_are_neighbors_and_not_listed_twice(
        self,
    ):
        """
        A-----------B-----------C
        |           |           |
        |           |           |
        |     D-----E-----F     |
        |     |           |     |
        |     |           |     |
        |     G-----H-----I     |
        |           |           |
        |           |           |
        J-----------K-----------L
        """
        topology = {
            "type": "Topology",
            "objects": {
                "abdeghkja": {"type": "Polygon", "arcs": [[0, 1, 2, 3]]},
                "bclkhifeb": {"type": "Polygon", "arcs": [[4, -3, 5, -1]]},
            },
            "arcs": [
                [[2, 0], [2, 1]],
                [[2, 1], [1, 1], [1, 2], [2, 2]],
                [[2, 2], [2, 3]],
                [[2, 3], [0, 3], [0, 0], [2, 0]],
                [[2, 0], [4, 0], [4, 3], [2, 3]],
                [[2, 2], [3, 2], [3, 1], [2, 1]],
            ],
        }
        self.assertListEqual(
            self.neighbors(
                [topology["objects"]["abdeghkja"], topology["objects"]["bclkhifeb"]]
            ),
            [[1], [0]],
        )



import json
import subprocess
import unittest


class TopoQuantizeTestCase(unittest.TestCase):
    def topoquantize(self, command, result):
        command = " ".join(command.split()).split(" ")
        p = subprocess.run(
            ["python", "bin/topoquantize.py", *command], capture_output=True
        )
        out, err = p.stdout, p.stderr
        actual = json.loads(out)
        with open(result, "r") as f:
            expected = json.load(f)

        self.assertDictEqual(actual, expected)

    def test_polygon_quantization(self):
        # self.topoquantize('-q 1e4 < tests/client/topojson/polygon.json',
        self.topoquantize(
            "-q 1e4 -i tests/client/topojson/polygon.json",
            "tests/client/topojson/polygon-q1e4.json",
        )

    def test_polygon_quantization_2(self):
        # self.topoquantize('-q 1e5 < tests/client/topojson/polygon.json',
        self.topoquantize(
            "-q 1e5 -i tests/client/topojson/polygon.json",
            "tests/client/topojson/polygon-q1e5.json",
        )

    def test_polygon_mercator_quantization(self):
        # self.topoquantize('-q 1e4 < tests/client/topojson/polygon-mercator.json',
        self.topoquantize(
            "-q 1e4 -i tests/client/topojson/polygon-mercator.json",
            "tests/client/topojson/polygon-mercator-q1e4.json",
        )

    def test_polygon_mercator_quantization_2(self):
        # self.topoquantize('-q 1e5 < tests/client/topojson/polygon-mercator.json',
        self.topoquantize(
            "-q 1e5 -i tests/client/topojson/polygon-mercator.json",
            "tests/client/topojson/polygon-mercator-q1e5.json",
        )

    def test_point_quantization(self):
        # self.topoquantize('-q 1e5 < tests/client/topojson/point.json',
        self.topoquantize(
            "-q 1e5 -i tests/client/topojson/point.json",
            "tests/client/topojson/point-q1e5.json",
        )

    def test_points_quantization(self):
        # self.topoquantize('-q 1e5 < tests/client/topojson/points.json',
        self.topoquantize(
            "-q 1e5 -i tests/client/topojson/points.json",
            "tests/client/topojson/points-q1e5.json",
        )



# Computes the bounding box of the specified hash of GeoJSON objects.
class BoundingBox(object):
    def __init__(self):
        self.x_0 = self.y_0 = float("inf")
        self.x_1 = self.y_1 = -float("inf")

    def __call__(self, geometry, *args, **kwargs):
        self.bound_geometry_type = {
            "GeometryCollection": lambda o: list(
                map(self.bound_geometry, o["geometries"])
            ),
            "Point": lambda o: self.bound_point(o["coordinates"]),
            "MultiPoint": lambda o: list(map(self.bound_point, o["coordinates"])),
            "LineString": lambda o: self.bound_line(o["arcs"]),
            "MultiLineString": lambda o: list(map(self.bound_line, o["arcs"])),
            "Polygon": lambda o: list(map(self.bound_line, o["arcs"])),
            "MultiPolygon": lambda o: list(map(self.bound_multiline, o["arcs"])),
        }

        for k in geometry:
            self.bound_geometry(geometry[k])

        self.value = (
            [self.x_0, self.y_0, self.x_1, self.y_1]
            if self.x_0 <= self.x_1 and self.y_0 <= self.y_1
            else None
        )
        return self.value

    def bound_geometry(self, geometry):
        if geometry and geometry["type"] in self.bound_geometry_type:
            self.bound_geometry_type[geometry["type"]](geometry)

    def bound_point(self, coordinates):
        x, y = coordinates
        self.x_0 = min(x, self.x_0)
        self.y_0 = min(y, self.y_0)
        self.x_1 = max(x, self.x_1)
        self.y_1 = max(y, self.y_1)

    def bound_line(self, coordinates):
        list(map(self.bound_point, coordinates))

    def bound_multiline(self, coordinates):
        list(map(self.bound_line, coordinates))



class Extract(object):
    """
    Extracts the lines and rings from the specified hash of geometry objects.

    Returns an object with three properties:

    * coordinates - shared buffer of [x, y] coordinates
    * lines - lines extracted from the hash, of the form [start, end]
    * rings - rings extracted from the hash, of the form [start, end]

    For each ring or line, start and end represent inclusive indexes into the
    coordinates buffer. For rings (and closed lines), coordinates[start] equals
    coordinates[end].

    For each line or polygon geometry in the input hash, including nested
    geometries as in geometry collections, the `coordinates` array is replaced
    with an equivalent `arcs` array that, for each line (for line string
    geometries) or ring (for polygon geometries), points to one of the above
    lines or rings.
    """

    def __init__(self):
        pass

    def __call__(self, objects, *args, **kwargs):
        self.extract_geometry_type = {
            "GeometryCollection": self._geometry_collection_call,
            "LineString": self._line_string_call,
            "MultiLineString": self._multi_line_string_call,
            "Polygon": self._polygon_call,
            "MultiPolygon": self._multi_polygon_call,
        }

        self.index = 0
        self.lines = list()
        self.rings = list()
        self.coordinates = list()

        for k in objects:
            self.extract_geometry(objects[k])

        return {
            "type": "Topology",
            "coordinates": self.coordinates,
            "lines": self.lines,
            "rings": self.rings,
            "objects": objects,
        }

    def _line_string_call(self, o):
        o["arcs"] = self.extract_line(o["arcs"])

    def _multi_line_string_call(self, o):
        o["arcs"] = list(map(self.extract_line, o["arcs"]))

    def _polygon_call(self, o):
        o["arcs"] = list(map(self.extract_ring, o["arcs"]))

    def _multi_polygon_call(self, o):
        o["arcs"] = list(map(self.extract_multi_ring, o["arcs"]))

    def _geometry_collection_call(self, o):
        for geometry in o["geometries"]:
            self.extract_geometry(geometry)

    def extract_geometry(self, geometry):
        if geometry and geometry["type"] in self.extract_geometry_type:
            self.extract_geometry_type[geometry["type"]](geometry)

    def extract(self, geometry, append_to):
        i = 0
        n = len(geometry)

        while i < n:
            self.coordinates.append(geometry[i])
            self.index += 1
            i += 1

        arc = [self.index - n, self.index - 1]

        append_to.append(arc)

        return arc

    def extract_line(self, line):
        return self.extract(line, self.lines)

    def extract_ring(self, ring):
        return self.extract(ring, self.rings)

    def extract_multi_ring(self, rings):
        return list(map(self.extract_ring, rings))



from pytopojson import feature, stitch


def planar_ring_area(ring):
    i = 0
    n = len(ring)
    b = ring[n - 1]
    area = 0
    while i < n:
        a, b = b, ring[i]
        area += a[0] * b[1] - a[1] * b[0]
        i += 1
    return abs(area)  # Note: Doubled area!


class MergeArcs(object):
    def __init__(self):
        self.object = feature.Object()
        self.stitch = stitch.Stitch()
        self.polygons_by_arc = dict()
        self.polygons = list()
        self.groups = list()
        self.topology = None

    def __call__(self, topology, objects, *args, **kwargs):
        self.topology = topology

        for o in objects:
            self.geometry(o)

        for polygon in self.polygons:
            if "_" not in polygon:
                group = list()
                polygon["_"] = 1
                neighbors = [polygon]
                polygon = neighbors.pop() if any(neighbors) else False

                while polygon:
                    group.append(polygon)
                    # DictList to list
                    p = list()
                    for i in range(len(polygon) - 1):
                        p.append(polygon[i])
                    for ring in p:
                        for arc in ring:
                            arc = ~arc if arc < 0 else arc
                            for p in self.polygons_by_arc[arc]:
                                if "_" not in p:
                                    p["_"] = 1
                                    neighbors.append(p)

                    polygon = neighbors.pop() if any(neighbors) else False

                for idx, point in enumerate(group):
                    if 1 in point:
                        group[idx] = [point[0], point[1]]
                    else:
                        group[idx] = [point[0]]
                self.groups.append(group)

        for idx, polygon in enumerate(self.polygons):
            if 1 in polygon:
                self.polygons[idx] = [polygon[0], polygon[1]]
            else:
                self.polygons[idx] = [polygon[0]]

        for i, _ in self.polygons_by_arc.items():
            for ii, _ in enumerate(self.polygons_by_arc[i]):
                self.polygons_by_arc[i][ii].pop("_", None)

        return {
            "type": "MultiPolygon",
            "arcs": list(map(lambda x: self._tmp(x), self.groups)),
        }

    def geometry(self, o):
        _type = o["type"]
        if _type == "GeometryCollection":
            for arc in o["geometries"]:
                self.geometry(arc)
        elif _type == "Polygon":
            self.extract(o["arcs"])
        elif _type == "MultiPolygon":
            for arc in o["arcs"]:
                self.extract(arc)

    def extract(self, polygon):
        p = dict()
        for idx, point in enumerate(polygon):
            p[idx] = point

        for ring in polygon:
            for arc in ring:
                arc = ~arc if arc < 0 else arc
                l = self.polygons_by_arc.get(arc, [])
                l.append(p)
                self.polygons_by_arc[arc] = l
        self.polygons.append(p)

    def area(self, ring):
        obj = self.object(self.topology, {"type": "Polygon", "arcs": [ring]})
        return planar_ring_area(obj["coordinates"][0])

    def _tmp(self, polygons):
        arcs = list()
        # Extract the exterior (unique) arcs
        for polygon in polygons:
            for ring in polygon:
                for arc in ring:
                    arc = ~arc if arc < 0 else arc
                    if len(self.polygons_by_arc[arc]) < 2:
                        arcs.append(arc)

        # Stitch the arcs into one or more rings
        arcs = self.stitch(self.topology, arcs)

        # If more than one ring is returned,
        # at most one of these rings can be the exterior;
        # choose the one with the greatest absolute area
        if 1 < len(arcs):
            k = self.area(arcs[0])
            for i in range(1, len(arcs)):
                k_i = self.area(arcs[i])
                if k < k_i:
                    arcs[0], arcs[i] = arcs[i], arcs[0]
                    k = k_i

        return list(filter(lambda x: 0 < len(x), arcs))


class Merge(object):
    def __init__(self):
        self.object = feature.Object()
        self.merge_arcs = MergeArcs()

    def __call__(self, topology, *args, **kwargs):
        arcs = self.merge_arcs(topology, *args)
        return self.object(topology, arcs)



class Bisect(object):
    def __init__(self):
        pass

    def __call__(self, a, x, *args, **kwargs):
        lo, hi = 0, len(a)

        while lo < hi:
            mid = lo + hi >> 1
            if a[mid] < x:
                lo = mid + 1
            else:
                hi = mid

        return lo



from pytopojson import transform


class Reverse(object):
    def __init__(self):
        pass

    def __call__(self, array, n, *args, **kwargs):
        j = len(array)
        i = j - n

        j -= 1
        while i < j:
            array[i], array[j] = array[j], array[i]
            i += 1
            j -= 1


class Object(object):
    def __init__(self):
        self.reverse = Reverse()
        self.transform = transform.Transform()
        self.transform_point = None
        self.arcs = None

    def __call__(self, topology, o, *args, **kwargs):
        self.transform_point = self.transform(topology.get("transform", None))
        self.arcs = topology["arcs"]
        return self.geometry(o)

    def arc(self, i, points):
        if any(points):
            points.pop()

        idx = ~i if i < 0 else i
        arcs = self.arcs[idx]
        for k, arc in enumerate(arcs):
            points.append(self.transform_point(arc, k))

        if i < 0:
            self.reverse(points, len(arcs))

    def point(self, p):
        return self.transform_point(p)

    def line(self, arcs):
        points = list()
        for arc in arcs:
            self.arc(arc, points)

        # This should never happen per the specification.
        if len(points) < 2:
            points.append(points[0])

        return points

    def ring(self, arcs):
        points = self.line(arcs)
        # This may happen if an arc has only two points.
        while len(points) < 4:
            points.append(points[0])
        return points

    def polygon(self, arcs):
        return list(map(lambda x: self.ring(x), arcs))

    def geometry(self, o):
        _type = o.get("type", None)

        if _type == "GeometryCollection":
            return {
                "type": _type,
                "geometries": list(map(lambda x: self.geometry(x), o["geometries"])),
            }
        elif _type == "Point":
            coordinates = self.point(o["coordinates"])
        elif _type == "MultiPoint":
            coordinates = list(map(lambda x: self.point(x), o["coordinates"]))
        elif _type == "LineString":
            coordinates = self.line(o["arcs"])
        elif _type == "MultiLineString":
            coordinates = list(map(lambda x: self.line(x), o["arcs"]))
        elif _type == "Polygon":
            coordinates = self.polygon(o["arcs"])
        elif _type == "MultiPolygon":
            coordinates = list(map(lambda x: self.polygon(x), o["arcs"]))
        else:
            return None

        return {"type": _type, "coordinates": coordinates}


class Feature(object):
    def __init__(self):
        self.object = Object()
        self.reverse = Reverse()
        self.transform = transform.Transform()

    def __call__(self, topology, o, *args, **kwargs):
        if isinstance(o, str):
            o = topology["objects"][o]
        if o.get("type", None) == "GeometryCollection":
            return {
                "type": "FeatureCollection",
                "features": list(
                    map(lambda x: self.feature(topology, x), o["geometries"])
                ),
            }
        else:
            return self.feature(topology, o)

    def feature(self, topology, o):
        feat = {
            "id": o.get("id", None),
            "bbox": o.get("bbox", None),
            "type": "Feature",
            "properties": o.get("properties", dict()),
            "geometry": self.object(topology, o),
        }

        for k in ("id", "bbox"):
            if feat[k] is None:
                del feat[k]

        return feat
